#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `buffer_work_space` module."""

import io
import os
import tempfile
import pytest
from pytest_mock import MockerFixture
from file_read_backwards.buffer_work_space import BufferWorkSpace
from file_read_backwards.buffer_work_space import new_lines_bytes
from file_read_backwards.buffer_work_space import _find_furthest_new_line
from file_read_backwards.buffer_work_space import _remove_trailing_new_line
from file_read_backwards.buffer_work_space import _get_file_size
from file_read_backwards.buffer_work_space import _is_partially_read_new_line
from file_read_backwards.buffer_work_space import _get_what_to_read_next
from file_read_backwards.buffer_work_space import _get_next_chunk


class TestFindFurthestNewLine:
    def test_find_furthest_new_line_with_no_new_line_in_empty_byte_string(self):
        test_string = b""
        r = _find_furthest_new_line(test_string)
        assert r == -1

    def test_find_furthest_new_line_with_no_new_line_in_non_empty_byte_string(self):
        test_string = b"SomeRandomCharacters"
        r = _find_furthest_new_line(test_string)
        assert r == -1

    def test_find_furthest_new_line_with_bytestring_with_new_line_at_the_end(self):
        base_string = b"SomeRandomCharacters"
        for n in new_lines_bytes:
            test_string = base_string + n
            expected_value = len(test_string) - 1
            r = _find_furthest_new_line(test_string)
            assert r == expected_value

    def test_find_furthest_new_line_with_bytestring_with_new_line_in_the_middle(self):
        base_string = b"SomeRandomCharacters"
        for n in new_lines_bytes:
            test_string = base_string + n + base_string
            expected_value = len(base_string) + len(n) - 1
            r = _find_furthest_new_line(test_string)
            assert r == expected_value

    def test_find_furthest_new_line_with_bytestring_with_new_line_in_the_middle_and_end(self):
        base_string = b"SomeRandomCharacters"
        for n in new_lines_bytes:
            test_string = base_string + n + base_string + n
            expected_value = len(test_string) - 1
            r = _find_furthest_new_line(test_string)
            assert r == expected_value


class TestRemoveTrailingNewLine:
    def test_remove_trailing_new_line_with_empty_byte_string(self):
        test_string = b""
        expected_string = test_string
        r = _remove_trailing_new_line(test_string)
        assert r == expected_string

    def test_remove_trailing_new_line_with_non_empty_byte_string_with_no_new_line(self):
        test_string = b"Something"
        expected_string = test_string
        r = _remove_trailing_new_line(test_string)
        assert r == expected_string

    def test_remove_trailing_new_line_with_non_empty_byte_string_with_variety_of_new_lines(self):
        expected_str = b"Something"
        for n in new_lines_bytes:
            test_string = expected_str + n
            r = _remove_trailing_new_line(test_string)
            assert r == expected_str

    def test_remove_trailing_new_line_with_non_empty_byte_string_with_variety_of_new_lines_in_the_middle(self):
        base_string = b"Something"
        for n in new_lines_bytes:
            test_string = base_string + n + base_string
            expected_string = test_string
            r = _remove_trailing_new_line(test_string)
            assert r == expected_string


class TestGetFileSize:
    def test_empty_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            pass
        expected_value = 0
        with io.open(t.name, mode="rb") as fp:
            r = _get_file_size(fp)
            assert r == expected_value
        os.unlink(t.name)

    def test_file_with_eight_bytes(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            t.write(b"a" * 8)
        expected_value = 8
        with io.open(t.name, mode="rb") as fp:
            r = _get_file_size(fp)
            assert r == expected_value
        os.unlink(t.name)


class TestIsPartiallyReadNewLine:
    def test_when_we_have_a_partially_read_new_line(self):
        for n in new_lines_bytes:
            if len(n) > 1:
                b = n[-1]
                r = _is_partially_read_new_line(b)
                assert r


class TestGetWhatToReadNext:
    def test_with_empty_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            pass
        expected_result = (0, 0)
        with io.open(t.name, mode="rb") as fp:
            r = _get_what_to_read_next(fp, previously_read_position=0, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)

    def test_with_file_with_seven_bytes_of_alphanumeric(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            t.write(b"abcdefg")
        expected_result = (4, 3)
        with io.open(t.name, mode="rb") as fp:
            r = _get_what_to_read_next(fp, previously_read_position=7, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)

    def test_with_file_with_single_new_line(self):
        for n in new_lines_bytes:
            with tempfile.NamedTemporaryFile(delete=False) as t:
                t.write(n)
            expected_result = (0, len(n))
            chunk_size = len(n) + 1
            with io.open(t.name, mode="rb") as fp:
                r = _get_what_to_read_next(fp, previously_read_position=len(n), chunk_size=chunk_size)
                assert r == expected_result
            os.unlink(t.name)

    def test_with_file_where_we_need_to_read_more_than_chunk_size(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            t.write(b"abcd\nfg")
        expected_result = (3, 4)
        with io.open(t.name, mode="rb") as fp:
            r = _get_what_to_read_next(fp, previously_read_position=7, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)


class TestGetNextChunk:
    def test_with_empty_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            pass
        expected_result = (b"", 0)
        with io.open(t.name, mode="rb") as fp:
            r = _get_next_chunk(fp, previously_read_position=0, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)

    def test_with_non_empty_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            t.write(b"abcdefg")
        expected_result = (b"efg", 4)
        with io.open(t.name, mode="rb") as fp:
            r = _get_next_chunk(fp, previously_read_position=7, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)

    def test_with_non_empty_file_where_we_read_more_than_chunk_size(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            t.write(b"abcd\nfg")
        expected_result = (b"d\nfg", 3)
        with io.open(t.name, mode="rb") as fp:
            r = _get_next_chunk(fp, previously_read_position=7, chunk_size=3)
            assert r == expected_result
        os.unlink(t.name)


class TestBufferWorkSpace:
    def test_add_to_empty_buffer_work_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024
        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        b.add_to_buffer(content=b"aaa", read_position=1021)
        assert b.read_buffer == b"aaa"
        assert b.read_position == 1021

    def test_add_to_non_empty_buffer_work_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024
        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        b.add_to_buffer(content=b"aaa", read_position=1021)
        b.add_to_buffer(content=b"bbb", read_position=1018)
        assert b.read_buffer == b"bbbaaa"
        assert b.read_position == 1018

    def test_yieldable_for_new_initialized_buffer_work_space(self):
        with tempfile.NamedTemporaryFile(delete=False) as t:
            with io.open(t.name, mode="rb") as fp:
                b = BufferWorkSpace(fp, chunk_size=io.DEFAULT_BUFFER_SIZE)
                r = b.yieldable()
                assert not r
        os.unlink(t.name)

    def test_yieldable_for_unexhausted_buffer_space_with_single_new_line(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 1024 - len(n)
            b.read_buffer = n
            expected_result = False
            r = b.yieldable()
            assert r == expected_result

    def test_yieldable_for_buffer_space_with_two_new_lines(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 1024 - (len(n) * 2)
            b.read_buffer = n * 2
            expected_result = True
            r = b.yieldable()
            assert r == expected_result

    def test_yieldable_for_fully_read_with_unreturned_contents_in_buffer_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 0
            b.read_buffer = b""
            expected_result = True
            r = b.yieldable()
            assert r == expected_result

    def test_yieldable_for_fully_read_and_returned_contents_in_buffer_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 0
            b.read_buffer = None
            r = b.yieldable()
            assert not r

    def test_return_line_with_buffer_space_with_two_new_lines(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 1024 - (len(n) * 2)
            b.read_buffer = n * 2
            expected_result = b""
            r = b.return_line()
            assert r == expected_result

    def test_return_line_with_buffer_space_with_some_contents_between_two_new_lines(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 1024 - (len(n) * 2)
            b.read_buffer = n + b"Something" + n
            expected_result = b"Something"
            r = b.return_line()
            assert r == expected_result

    def test_return_line_with_buffer_space_with_fully_read_in_contents_at_its_last_line(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        for n in new_lines_bytes:
            b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
            b.read_position = 0
            b.read_buffer = b"LastLineYay"
            expected_result = b"LastLineYay"
            r = b.return_line()
            assert r == expected_result

    def test_return_line_contract_violation(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 0

        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        with pytest.raises(AssertionError):
            b.return_line()

    def test_has_returned_every_line_empty_file(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 0

        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        r = b.has_returned_every_line()
        assert r

    def test_has_returned_every_line_with_not_fully_read_in_buffer_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        b.read_position = 1
        r = b.has_returned_every_line()
        assert not r

    def test_has_returned_every_line_with_fully_read_in_and_unprocessed_buffer_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        b.read_position = 0
        b.read_buffer = b"abc"
        r = b.has_returned_every_line()
        assert not r

    def test_has_returned_every_line_with_fully_read_in_and_processed_buffer_space(self, mocker: MockerFixture):
        _get_file_size_mock = mocker.patch("file_read_backwards.buffer_work_space._get_file_size")
        fp_mock = mocker.Mock()
        _get_file_size_mock.return_value = 1024

        b = BufferWorkSpace(fp_mock, chunk_size=io.DEFAULT_BUFFER_SIZE)
        b.read_position = 0
        b.read_buffer = None
        r = b.has_returned_every_line()
        assert r



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `file_read_backwards` module."""

import itertools
import os
import tempfile
import pytest

from collections import deque

from file_read_backwards.file_read_backwards import FileReadBackwards
from file_read_backwards.file_read_backwards import supported_encodings
from file_read_backwards.buffer_work_space import new_lines


# doing this xrange/range dance so that we don't need to add additional dependencies of future or six modules
try:
    xrange
except NameError:
    xrange = range


created_files = set()


def helper_write(t, s, encoding="utf-8"):
    """A helper method to write out string s in specified encoding."""
    t.write(s.encode(encoding))


def helper_create_temp_file(generator=None, encoding='utf-8'):
    global created_files
    if generator is None:
        generator = ("line {}!\n".format(i) for i in xrange(42))
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    for line in generator:
        helper_write(temp_file, line, encoding)
    temp_file.close()
    print('Wrote file {}'.format(temp_file.name))
    created_files.add(temp_file)
    return temp_file


def helper_destroy_temp_file(temp_file):
    temp_file.close()
    os.unlink(temp_file.name)


def helper_destroy_temp_files():
    global created_files
    while created_files:
        helper_destroy_temp_file(created_files.pop())


@pytest.fixture(scope="module")
def empty_file():
    return helper_create_temp_file(generator=(_ for _ in []))


@pytest.fixture(scope="module")
def long_file():
    return helper_create_temp_file()


@pytest.fixture(scope="module", autouse=True)
def cleanup_files():
    yield
    helper_destroy_temp_files()


class TestFileReadBackwards:
    def test_with_completely_empty_file(self, empty_file):
        f = FileReadBackwards(empty_file.name)
        expected_lines = deque()
        lines_read = deque()
        for line in f:
            lines_read.appendleft(line)
        assert expected_lines == lines_read

    def test_file_with_a_single_new_line_char_with_different_encodings(self):
        for encoding, new_line in itertools.product(supported_encodings, new_lines):
            temp_file = helper_create_temp_file((line for line in [new_line]), encoding=encoding)
            f = FileReadBackwards(temp_file.name)
            expected_lines = deque([""])
            lines_read = deque()
            for line in f:
                lines_read.appendleft(line)
            assert expected_lines == lines_read

    def test_file_with_one_line_of_text_with_accented_char_followed_by_a_new_line(self):
        b = b'Caf\xc3\xa9'  # accented e in utf-8
        s = b.decode("utf-8")
        for new_line in new_lines:
            temp_file = helper_create_temp_file((line for line in [s, new_line]))
            f = FileReadBackwards(temp_file.name)
            expected_lines = deque([s])
            lines_read = deque()
            for line in f:
                lines_read.appendleft(s)
            assert expected_lines == lines_read

    def test_file_with_one_line_of_text_followed_by_a_new_line_with_different_encodings(self):
        for encoding, new_line in itertools.product(supported_encodings, new_lines):
            temp_file = helper_create_temp_file((line for line in ["something{0}".format(new_line)]), encoding=encoding)
            f = FileReadBackwards(temp_file.name)
            expected_lines = deque(["something"])
            lines_read = deque()
            for line in f:
                lines_read.appendleft(line)
            assert expected_lines == lines_read

    def test_file_with_varying_number_of_new_lines_and_some_text_in_chunk_size(self):
        chunk_size = 3
        s = "t"
        for number_of_new_lines in xrange(21):
            for new_line in new_lines:
                temp_file = helper_create_temp_file((line for line in [new_line * number_of_new_lines, s * chunk_size]))
                f = FileReadBackwards(temp_file.name, chunk_size=chunk_size)
                expected_lines = deque()
                for _ in xrange(number_of_new_lines):
                    expected_lines.append("")
                expected_lines.append(s * chunk_size)
                lines_read = deque()
                for line in f:
                    lines_read.appendleft(line)
                assert expected_lines == lines_read

    def test_file_with_new_lines_and_some_accented_characters_in_chunk_size(self):
        chunk_size = 3
        b = b'\xc3\xa9'
        s = b.decode("utf-8")
        for number_of_new_lines in xrange(21):
            for new_line in new_lines:
                temp_file = helper_create_temp_file((line for line in [new_line * number_of_new_lines, s * chunk_size]))
                f = FileReadBackwards(temp_file.name, chunk_size=chunk_size)
                expected_lines = deque()
                for _ in xrange(number_of_new_lines):
                    expected_lines.append("")
                expected_lines.append(s * chunk_size)
                lines_read = deque()
                for line in f:
                    lines_read.appendleft(line)
                assert expected_lines == lines_read

    def test_unsupported_encoding(self, empty_file):
        with pytest.raises(NotImplementedError):
            _ = FileReadBackwards(empty_file.name, encoding="not-supported-encoding")

    def test_file_with_one_line_of_text_readline(self):
        s = "Line0"
        for new_line in new_lines:
            temp_file = helper_create_temp_file((line for line in [s, new_line]))
            with FileReadBackwards(temp_file.name) as fp:
                line = fp.readline()
                expected_line = s + os.linesep
                assert line == expected_line

                second_line = fp.readline()
                expected_second_line = ""
                assert second_line == expected_second_line

    def test_file_with_two_lines_of_text_readline(self):
        line0 = "Line0"
        line1 = "Line1"
        for new_line in new_lines:
            line0_with_n = "{}{}".format(line0, new_line)
            line1_with_n = "{}{}".format(line1, new_line)
            temp_file = helper_create_temp_file((line for line in [line0_with_n, line1_with_n]))
            with FileReadBackwards(temp_file.name) as fp:
                line = fp.readline()
                expected_line = line1 + os.linesep
                assert line == expected_line

                second_line = fp.readline()
                expected_second_line = line0 + os.linesep
                assert second_line == expected_second_line

                third_line = fp.readline()
                expected_third_line = ""
                assert third_line == expected_third_line


class TestFileReadBackwardsAsContextManager:
    @pytest.fixture(scope="class", autouse=True)
    def temp_file(self):
        return helper_create_temp_file()

    def test_behaves_as_classic(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            lines_read = deque()
            for line in f:
                lines_read.appendleft(line)
        f2 = FileReadBackwards(temp_file.name)
        lines_read2 = deque()
        for l2 in f2:
            lines_read2.appendleft(l2)
        assert lines_read == lines_read2


class TestFileReadBackwardsCloseFunctionality:
    @pytest.fixture(scope="class", autouse=True)
    def temp_file(self):
        return helper_create_temp_file()

    def test_close_on_iterator(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it = iter(f)
            for count, i in enumerate(it):
                if count == 2:
                    break
            assert not it.closed
            it.close()
            assert it.closed

    def test_not_creating_new_iterator(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it1 = iter(f)
            it2 = iter(f)
            assert it1 is it2

    def test_close_on_iterator_exhausted(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it = iter(f)
            for _ in it:
                pass
            assert it.closed

    def test_close_on_reader_exit(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it = iter(f)
        assert it.closed

    def test_close_on_reader_explicitly(self, temp_file):
        f = FileReadBackwards(temp_file.name)
        it = iter(f)
        assert not it.closed
        f.close()
        assert it.closed

    def test_close_on_reader_with_already_closed_iterator(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it = iter(f)
            it.close()
        assert it.closed

    def test_cannot_iterate_when_closed(self, temp_file):
        with FileReadBackwards(temp_file.name) as f:
            it = iter(f)
            it.close()
            for _ in it:
                pytest.fail("An iterator should be exhausted when closed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
]

test_requirements = [
    "mock",
]

setup(
    name='file_read_backwards',
    version='3.2.0',
    description="Memory efficient way of reading files line-by-line from the end of file",
    long_description=readme + '\n\n' + history,
    author="Robin Robin",
    author_email='robinsquare42@gmail.com',
    url='https://github.com/RobinNil/file_read_backwards',
    packages=[
        'file_read_backwards',
    ],
    package_dir={'file_read_backwards':
                 'file_read_backwards'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='file_read_backwards',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
